#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#    __init__.py.py
#      Version:  1.0
#      Goal: Class file / Test script
#      Created by @wenchieh  on <12/17/2017>


__author__ = 'wenchieh'
