#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Configure names exposed to spartan.
'''

# here put the import lib

from .backend import load_backend, get_backend
from .task import *
from .util import *
from .model import *
from .tensor import *
